package de.ozgcloud.bescheid;

import static de.ozgcloud.bescheid.BescheidEventListener.*;
import static de.ozgcloud.bescheid.BescheidRequestTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.HashMap;
import java.util.Map;
import java.util.Optional;
import java.util.function.Consumer;

import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.security.core.context.SecurityContext;

import de.ozgcloud.bescheid.Bescheid.SendBy;
import de.ozgcloud.bescheid.binaryfile.BinaryFileService;
import de.ozgcloud.bescheid.common.callcontext.CurrentUserService;
import de.ozgcloud.bescheid.common.freemarker.TemplateHandler;
import de.ozgcloud.bescheid.common.user.UserProfile;
import de.ozgcloud.bescheid.common.user.UserProfileService;
import de.ozgcloud.bescheid.common.user.UserProfileTestFactory;
import de.ozgcloud.command.Command;
import de.ozgcloud.command.CommandCreatedEventTestFactory;
import de.ozgcloud.command.CommandExecutedEvent;
import de.ozgcloud.command.CommandFailedEvent;
import de.ozgcloud.command.CommandTestFactory;
import de.ozgcloud.common.attached_item.AttachedItem;
import de.ozgcloud.common.attached_item.AttachedItemService;
import de.ozgcloud.common.attached_item.AttachedItemTestFactory;
import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.document.BescheidDocumentCreatedEvent;
import de.ozgcloud.document.Document;
import de.ozgcloud.document.DocumentService;
import de.ozgcloud.document.DocumentTestFactory;

class BescheidEventListenerTest {

	@Spy
	@InjectMocks
	private BescheidEventListener listener;

	@Mock
	private BescheidService service;
	@Mock
	private BinaryFileService fileService;
	@Mock
	private AttachedItemService attachedItemService;
	@Mock
	private DocumentService documentService;

	@Mock
	private ApplicationEventPublisher eventPublisher;
	@Mock
	private CurrentUserService userService;
	@Mock
	private UserProfileService userProfileService;
	@Mock
	private TemplateHandler templateHandler;

	@Nested
	class TestOnCreateBescheidCommand {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallRunWithSecurityContext() {
			listener.onCreateBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).runWithSecurityContext(eq(command), any());
		}

		@Test
		void shouldExecuteCreateBescheid() {
			listener.onCreateBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).doCreateBescheid(command);
		}
	}

	@Nested
	class TestDoCreateBescheid {

		private static final Command COMMAND = CommandTestFactory.createBuilder()
				.bodyObject(
						Map.of(VORGANG_ID_BODY_KEY, VORGANG_ID.toString(),
								BESCHEID_VOM_BODY_KEY, BESCHEID_VOM_STRING,
								GENEHMIGT_BODY_KEY, GENEHMIGT))
				.build();

		@Captor
		private ArgumentCaptor<BescheidCreatedEvent> eventCaptor;

		@Test
		void shouldCallCreateBescheidDraft() {
			listener.doCreateBescheid(COMMAND);

			verify(service).createBescheidDraft(COMMAND);
		}

		@Test
		@DisplayName("should publish BescheidCreatedEvent after creating BescheidDraft")
		void shouldPublishBescheidCreatedEventWithCommand() {
			var createdResource = "item-id";
			when(service.createBescheidDraft(any())).thenReturn(createdResource);

			listener.doCreateBescheid(COMMAND);

			verify(eventPublisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue().getCommand()).isSameAs(COMMAND);
		}

		@Test
		@DisplayName("should publish BescheidCreatedEvent with created resource")
		void shouldPublishBescheidCreatedEventWithCreatedResource() {
			var createdResource = "item-id";
			when(service.createBescheidDraft(any())).thenReturn(createdResource);

			listener.doCreateBescheid(COMMAND);

			verify(eventPublisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue().getCreatedResource()).isEqualTo(createdResource);
		}
	}

	@Nested
	class TestCreateBescheidRequest {

		private static final Command COMMAND = CommandTestFactory.createBuilder()
				.vorgangId(VORGANG_ID.toString())
				.bodyObject(
						Map.of(BESCHEID_VOM_BODY_KEY, BESCHEID_VOM_STRING,
								GENEHMIGT_BODY_KEY, false))
				.build();

		@Test
		void shouldSetVorgangId() {
			var request = createBescheidRequest();

			assertThat(request.getVorgangId()).isEqualTo(VORGANG_ID);
		}

		@Test
		void shouldSetBescheidVom() {
			var request = createBescheidRequest();

			assertThat(request.getBescheidVom()).isEqualTo(BESCHEID_VOM);
		}

		@Test
		void shouldSetGenehmigt() {
			var request = createBescheidRequest();

			assertThat(request.isGenehmigt()).isFalse();
		}

		@Test
		void shouldSetDefaultGenehmigt() {
			Command command = CommandTestFactory.createBuilder()
					.vorgangId(VORGANG_ID.toString())
					.bodyObject(
							Map.of(BESCHEID_VOM_BODY_KEY, BESCHEID_VOM_STRING))
					.build();

			var request = listener.createRequest(command);

			assertThat(request.isGenehmigt()).isTrue();
		}

		@Test
		void shouldContainUserProfile() {
			UserProfile user = UserProfileTestFactory.create();
			when(userProfileService.getUserProfile()).thenReturn(user);

			var request = listener.createRequest(COMMAND);

			assertThat(request.getCreateFor()).isSameAs(user);
		}

		private BescheidRequest createBescheidRequest() {
			return listener.createRequest(COMMAND);
		}
	}

	@Nested
	class TestOnDeleteBescheid {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallRunWithSecurityContext() {
			listener.onDeleteBescheid(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).runWithSecurityContext(eq(command), any());
		}

		@Test
		void shouldExecuteCreateBescheid() {
			listener.onDeleteBescheid(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).doDeleteBescheid(command);
		}
	}

	@Nested
	class TestDoDeleteBescheid {

		@Captor
		private ArgumentCaptor<BescheidDeletedEvent> bescheidDeletedEventCaptor;

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallAttachedItemService() {
			listener.doDeleteBescheid(command);

			verify(attachedItemService).deleteBescheidDraft(command);
		}

		@Test
		void shouldPublishCommandExecutedEvent() {
			listener.doDeleteBescheid(command);

			verify(eventPublisher).publishEvent(bescheidDeletedEventCaptor.capture());
			assertThat(bescheidDeletedEventCaptor.getValue().getCommand()).isSameAs(command);
		}

	}

	@Nested
	class TestOnUpdateBescheid {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallRunWithSecurityContext() {
			listener.onUpdateBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).runWithSecurityContext(eq(command), any());
		}

		@Test
		void shouldExecuteCreateBescheid() {
			listener.onUpdateBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).doUpdateBescheid(command);
		}

	}

	@Nested
	class TestDoUpdateBescheid {

		@Captor
		private ArgumentCaptor<BescheidUpdatedEvent> bescheidUpdatedEventCaptor;

		private final Command command = CommandTestFactory.createBuilder().build();
		private final Document document = DocumentTestFactory.create();

		@BeforeEach
		void mock() {
			doReturn(Optional.of(document)).when(listener).getDocument(command);
		}

		@Test
		void shouldGetDocument() {
			listener.doUpdateBescheid(command);

			verify(listener).getDocument(command);
		}

		@Test
		void shouldCallUpdateBescheidDraft() {
			listener.doUpdateBescheid(command);

			verify(attachedItemService).updateBescheidDraft(command, Optional.of(document));
		}

		@Test
		void shouldPublishBescheidUpdatedEvent() {
			listener.doUpdateBescheid(command);

			verify(eventPublisher).publishEvent(bescheidUpdatedEventCaptor.capture());
			assertThat(bescheidUpdatedEventCaptor.getValue().getCommand()).isSameAs(command);
		}
	}

	@DisplayName("Get document")
	@Nested
	class TestGetDocument {

		@DisplayName("if exists")
		@Nested
		class TestIfExists {

			private final Command command = CommandTestFactory.createBuilder().bodyObject(AttachedItemTestFactory.createBescheidItem()).build();

			@BeforeEach
			void mock() {
				when(documentService.getDocument(any())).thenReturn(DocumentTestFactory.create());
			}

			@Test
			void shouldCallService() {
				listener.getDocument(command);

				verify(documentService).getDocument(DocumentTestFactory.ID);
			}

			@Test
			void shouldReturnValue() {
				var document = listener.getDocument(command);

				assertThat(document.get()).usingRecursiveComparison().isEqualTo(DocumentTestFactory.create());
			}
		}

		@Test
		void shouldReturnEmptyIfNoDocumentExists() {
			var document = listener.getDocument(createCommandWithoutDocument());

			assertThat(document).isEmpty();
		}

		private Command createCommandWithoutDocument() {
			var bescheidItemMap = AttachedItemTestFactory.createBescheidItem();
			bescheidItemMap.remove(DocumentTestFactory.ID);
			return CommandTestFactory.createBuilder().bodyObject(bescheidItemMap).build();
		}
	}

	@Nested
	class TestOnCreateBescheidDocument {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallRunWithSecurityContext() {
			listener.onCreatedBescheidDocument(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).runWithSecurityContext(eq(command), any());
		}

		@Test
		void shouldExecuteCreateBescheidDocument() {
			listener.onCreatedBescheidDocument(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).doCreateBescheidDocument(command);
		}
	}

	@Nested
	class TestDoCreateBescheidDocument {

		private static final Command COMMAND = CommandTestFactory.create();
		private static final BescheidResponse BESCHEID = BescheidResponseTestFactory.create();

		@Captor
		private ArgumentCaptor<BescheidDocumentCreatedEvent> eventCaptor;

		@BeforeEach
		void init() {
			doReturn(BESCHEID).when(listener).doCreateBescheidBiz(any());
		}

		@Test
		void shouldCallCreateBescheidBiz() {
			listener.doCreateBescheidDocument(COMMAND);

			verify(listener).doCreateBescheidBiz(COMMAND);
		}

		@Test
		void shouldCallCreateBescheidDocument() {
			listener.doCreateBescheidDocument(COMMAND);

			verify(documentService).createBescheidDocument(COMMAND, BESCHEID);
		}

		@Test
		@DisplayName("should publish BescheidDocumentCreatedEvent with command")
		void shouldPublishEventWithCommand() {
			listener.doCreateBescheidDocument(COMMAND);

			verify(eventPublisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue().getCommand()).isEqualTo(COMMAND);
		}

		@Test
		@DisplayName("should publish BescheidDocumentCreatedEvent with createdResource")
		void shouldPublishEventWithCreatedResource() {
			var bescheidDocument = "document-id";
			when(documentService.createBescheidDocument(any(), any(BescheidResponse.class))).thenReturn(bescheidDocument);

			listener.doCreateBescheidDocument(COMMAND);

			verify(eventPublisher).publishEvent(eventCaptor.capture());
			assertThat(eventCaptor.getValue().getCreatedResource()).isEqualTo(bescheidDocument);
		}
	}

	@Nested
	class TestCreateBescheidBiz {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallCreateRequest() {
			listener.doCreateBescheidBiz(command);

			verify(listener).createRequest(command);
		}

		@Test
		void shouldCallBescheidService() {
			var bescheidRequest = BescheidRequestTestFactory.create();
			doReturn(bescheidRequest).when(listener).createRequest(any());

			listener.doCreateBescheidBiz(command);

			verify(service).createBescheid(bescheidRequest);
		}

		@Test
		void shouldCallFileService() {
			var bescheid = BescheidResponseTestFactory.create();
			when(service.createBescheid(any())).thenReturn(bescheid);

			listener.doCreateBescheidBiz(command);

			verify(fileService).uploadBescheidFile(bescheid);
		}

		@Test
		void shouldReturnBescheid() {
			var bescheid = BescheidResponseTestFactory.create();
			when(fileService.uploadBescheidFile(any())).thenReturn(bescheid);

			var result = listener.doCreateBescheidBiz(command);

			assertThat(result).isSameAs(bescheid);
		}
	}

	@Nested
	class TestOnSendBescheidCommand {

		private final Command command = CommandTestFactory.create();

		@Test
		void shouldCallRunWithSecurityContext() {
			listener.onSendBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).runWithSecurityContext(eq(command), any());
		}

		@Test
		void shouldExecuteDoSendBescheid() {
			listener.onSendBescheidCommand(CommandCreatedEventTestFactory.withCommand(command));

			verify(listener).doSendBescheid(command);
		}
	}

	@Nested
	class TestDoSendBescheid {

		@Captor
		private ArgumentCaptor<BescheidSentEvent> bescheidSentEventCaptor;

		@Test
		void shouldThrowExceptionIfUnkonwnSendBy() {
			var item = AttachedItemTestFactory.createBescheidItem();
			item.put(Bescheid.FIELD_SEND_BY, "unknown");
			var bescheidItem = AttachedItemTestFactory.createBescheidBuilder().item(item).build();
			when(attachedItemService.getItem(any())).thenReturn(bescheidItem);
			var command = CommandTestFactory.createBuilder().relationId(AttachedItemTestFactory.ID).build();

			Assertions.assertThrows(TechnicalException.class, () -> listener.doSendBescheid(command));
		}

		@DisplayName("manually")
		@Nested
		class TestManually {

			private static final Command COMMAND = CommandTestFactory.createBuilder().relationId(AttachedItemTestFactory.ID)
					.relationVersion(AttachedItemTestFactory.VERSION).bodyObject(AttachedItemTestFactory.createBescheidItem()).build();

			private final AttachedItem bescheidItem = createManuallyBescheidItem();

			private AttachedItem createManuallyBescheidItem() {
				var attachedItem = AttachedItemTestFactory.createBescheid();
				var item = new HashMap<>(attachedItem.getItem());

				item.put(Bescheid.FIELD_SEND_BY, SendBy.MANUAL);

				return attachedItem.toBuilder().item(item).build();
			}

			@BeforeEach
			void init() {
				when(attachedItemService.getItem(any())).thenReturn(bescheidItem);
			}

			@Test
			void shouldCallSendBescheid() {
				listener.doSendBescheid(COMMAND);

				verify(service).sendBescheidManually(bescheidItem, COMMAND);
				verify(service, never()).sendBescheidPostfachMail(any(), any());
			}

			@Test
			void shouldPublishEvent() {
				listener.doSendBescheid(COMMAND);

				verify(eventPublisher).publishEvent(bescheidSentEventCaptor.capture());
				assertThat(bescheidSentEventCaptor.getValue().getCommand()).isEqualTo(COMMAND);
			}
		}

		@DisplayName("per postfach mail")
		@Nested
		class TestPostfachMail {

			private static final Command COMMAND = CommandTestFactory.createBuilder().relationId(AttachedItemTestFactory.ID)
					.relationVersion(AttachedItemTestFactory.VERSION).bodyObject(AttachedItemTestFactory.createBescheidItem()).build();

			private AttachedItem bescheidItem;

			@BeforeEach
			void init() {
				var item = AttachedItemTestFactory.createBescheidItem();
				item.put(Bescheid.FIELD_SEND_BY, Bescheid.SendBy.NACHRICHT.name());
				bescheidItem = AttachedItemTestFactory.createBescheidBuilder().item(item).build();
				when(attachedItemService.getItem(any())).thenReturn(bescheidItem);
			}

			@Test
			void shouldCallBescheidService() {
				listener.doSendBescheid(COMMAND);

				verify(service).sendBescheidPostfachMail(bescheidItem, COMMAND);
				verify(service, never()).sendBescheidManually(any(), any());
			}

			@Test
			void shouldPublishEvent() {
				listener.doSendBescheid(COMMAND);

				verify(eventPublisher).publishEvent(bescheidSentEventCaptor.capture());
				assertThat(bescheidSentEventCaptor.getValue().getCommand()).isEqualTo(COMMAND);
			}
		}
	}

	@Nested
	class TestOnBescheidSent {

		private static final Command COMMAND = CommandTestFactory.create();

		@Test
		void shouldCallRunWithSecurityContext() {
			listener.onBescheidSent(new CommandExecutedEvent(COMMAND));

			verify(listener).runWithSecurityContext(eq(COMMAND), any());
		}

		@Test
		void shouldExecuteSetAntragBewilligung() {
			listener.onBescheidSent(new CommandExecutedEvent(COMMAND));

			verify(listener).setAntragBewilligung(COMMAND);
		}
	}

	@Nested
	class TestSetAntragBewilligung {

		private static final Command COMMAND = CommandTestFactory.create();

		@Test
		void shouldCallGetItem() {
			listener.setAntragBewilligung(COMMAND);

			verify(attachedItemService).getItem(CommandTestFactory.RELATION_ID);
		}

		@Test
		void shouldCallSetAntragBewilligung() {
			var item = AttachedItemTestFactory.createBescheid();
			when(attachedItemService.getItem(any())).thenReturn(item);

			listener.setAntragBewilligung(COMMAND);

			verify(service).setAntragBewilligung(item);
		}
	}

	@Nested
	class TestRunWithSecurityContext {

		private final Command command = CommandTestFactory.createBuilder().build();

		@Mock
		private Consumer<Command> commandExecutor;
		@Mock
		private SecurityContext secContext;
		@Captor
		private ArgumentCaptor<CommandFailedEvent> commandFailedEventCaptor;

		@BeforeEach
		void init() {
			when(userService.startSecurityContext(any())).thenReturn(secContext);
		}

		@Test
		void shouldStartSecurityContext() {
			listener.runWithSecurityContext(command, commandExecutor);

			verify(userService).startSecurityContext(command);
		}

		@Test
		void shouldExecuteCommand() {
			listener.runWithSecurityContext(command, commandExecutor);

			verify(commandExecutor).accept(command);
		}

		@Test
		void shouldResetSecurityContext() {
			listener.runWithSecurityContext(command, commandExecutor);

			verify(userService).resetSecurityContext(secContext);
		}

		@Test
		void shouldResetSecurityContextAfterException() {
			doThrow(new RuntimeException("ups")).when(commandExecutor).accept(any());

			listener.runWithSecurityContext(command, commandExecutor);

			verify(userService).resetSecurityContext(secContext);
		}

		@Test
		void shouldPublishCommandFailedEvent() {
			doThrow(new RuntimeException("ups")).when(commandExecutor).accept(any());

			listener.runWithSecurityContext(command, commandExecutor);

			verify(eventPublisher).publishEvent(commandFailedEventCaptor.capture());
			assertThat(commandFailedEventCaptor.getValue().getSource()).isEqualTo(command.getId());
			assertThat(commandFailedEventCaptor.getValue().getErrorMessage()).isNotEmpty();
		}
	}
}
